from dataclasses import dataclass
from time import time
from typing import Optional
from thirdweb.common.signature_minting import resolve_or_generate_id
from thirdweb.types.currency import Price
from thirdweb.constants.currency import NATIVE_TOKEN_ADDRESS, ZERO_ADDRESS
from thirdweb.types.nft import NFTMetadataInput


@dataclass
class BaseSignaturePayloadInput:
    to: str = ZERO_ADDRESS
    price: Price = 0
    currency_address: str = NATIVE_TOKEN_ADDRESS
    mint_start_time: int = int(time())
    mint_end_time: int = int(time()) + 60 * 60 * 24 * 1000 * 1000
    uid: Optional[bytes] = None
    primary_sale_recipient: Optional[str] = ZERO_ADDRESS

    def set_uid(self, uid: str) -> "BaseSignaturePayloadInput":
        self.uid = resolve_or_generate_id(uid)
        return self


@dataclass
class Signature20PayloadInput(BaseSignaturePayloadInput):
    quantity: Price = 0


@dataclass
class Signature20PayloadOutput(Signature20PayloadInput):
    pass


@dataclass
class Signature721PayloadInput:
    metadata: NFTMetadataInput
    royalty_recipient: str = ZERO_ADDRESS
    royalty_bps: int = 0
    to: str = ZERO_ADDRESS
    price: Price = 0
    currency_address: str = NATIVE_TOKEN_ADDRESS
    mint_start_time: int = int(time())
    mint_end_time: int = int(time()) + 60 * 60 * 24 * 1000 * 1000
    uid: Optional[bytes] = None
    primary_sale_recipient: Optional[str] = ZERO_ADDRESS

    def set_uid(self, uid: str) -> "Signature721PayloadInput":
        self.uid = resolve_or_generate_id(uid)
        return self


@dataclass
class Signature721PayloadOutput:
    metadata: NFTMetadataInput
    royalty_recipient: str = ZERO_ADDRESS
    royalty_bps: int = 0
    uri: str = ""
    to: str = ZERO_ADDRESS
    price: Price = 0
    currency_address: str = NATIVE_TOKEN_ADDRESS
    mint_start_time: int = int(time())
    mint_end_time: int = int(time()) + 60 * 60 * 24 * 1000 * 1000
    uid: Optional[bytes] = None
    primary_sale_recipient: Optional[str] = ZERO_ADDRESS

    def set_uid(self, uid: str) -> "Signature721PayloadOutput":
        self.uid = resolve_or_generate_id(uid)
        return self


@dataclass
class Signature1155PayloadInput:
    metadata: NFTMetadataInput
    token_id: int
    quantity: int
    royalty_recipient: str = ZERO_ADDRESS
    royalty_bps: int = 0
    to: str = ZERO_ADDRESS
    price: Price = 0
    currency_address: str = NATIVE_TOKEN_ADDRESS
    mint_start_time: int = int(time())
    mint_end_time: int = int(time()) + 60 * 60 * 24 * 1000 * 1000
    uid: Optional[bytes] = None
    primary_sale_recipient: Optional[str] = ZERO_ADDRESS

    def set_uid(self, uid: str) -> "Signature1155PayloadInput":
        self.uid = resolve_or_generate_id(uid)
        return self


@dataclass
class Signature1155PayloadOutput:
    metadata: NFTMetadataInput
    royalty_bps: int
    royalty_recipient: str
    uri: str
    token_id: int
    quantity: int
    to: str = ZERO_ADDRESS
    price: Price = 0
    currency_address: str = NATIVE_TOKEN_ADDRESS
    mint_start_time: int = int(time())
    mint_end_time: int = int(time()) + 60 * 60 * 24 * 1000 * 1000
    uid: Optional[bytes] = None
    primary_sale_recipient: Optional[str] = ZERO_ADDRESS

    def set_uid(self, uid: str) -> "Signature1155PayloadOutput":
        self.uid = resolve_or_generate_id(uid)
        return self


FilledSignaturePayload20 = Signature20PayloadInput
PayloadWithUri20 = Signature20PayloadOutput
PayloadToSign20 = Signature20PayloadInput


@dataclass
class SignedPayload20:
    payload: Signature20PayloadInput
    signature: str


FilledSignaturePayload721 = Signature721PayloadInput
PayloadWithUri721 = Signature721PayloadOutput
PayloadToSign721 = Signature721PayloadInput


@dataclass
class SignedPayload721:
    payload: Signature721PayloadInput
    signature: str


FilledSignaturePayload1155 = Signature1155PayloadInput
PayloadWithUri1155 = Signature1155PayloadOutput
PayloadToSign1155 = Signature1155PayloadInput


@dataclass
class SignedPayload1155:
    payload: Signature1155PayloadInput
    signature: str


MintRequest20 = [
    {"name": "to", "type": "address"},
    {"name": "primarySaleRecipient", "type": "address"},
    {"name": "quantity", "type": "uint256"},
    {"name": "price", "type": "uint256"},
    {"name": "currency", "type": "address"},
    {"name": "validityStartTimestamp", "type": "uint128"},
    {"name": "validityEndTimestamp", "type": "uint128"},
    {"name": "uid", "type": "bytes32"},
]

MintRequest721 = [
    {"name": "to", "type": "address"},
    {"name": "royaltyRecipient", "type": "address"},
    {"name": "royaltyBps", "type": "uint256"},
    {"name": "primarySaleRecipient", "type": "address"},
    {"name": "uri", "type": "string"},
    {"name": "price", "type": "uint256"},
    {"name": "currency", "type": "address"},
    {"name": "validityStartTimestamp", "type": "uint128"},
    {"name": "validityEndTimestamp", "type": "uint128"},
    {"name": "uid", "type": "bytes32"},
]

MintRequest1155 = [
    {"name": "to", "type": "address"},
    {"name": "royaltyRecipient", "type": "address"},
    {"name": "royaltyBps", "type": "uint256"},
    {"name": "primarySaleRecipient", "type": "address"},
    {"name": "tokenId", "type": "uint256"},
    {"name": "uri", "type": "string"},
    {"name": "quantity", "type": "uint256"},
    {"name": "pricePerToken", "type": "uint256"},
    {"name": "currency", "type": "address"},
    {"name": "validityStartTimestamp", "type": "uint128"},
    {"name": "validityEndTimestamp", "type": "uint128"},
    {"name": "uid", "type": "bytes32"},
]

EIP712DomainType = [
    {"name": "name", "type": "string"},
    {"name": "version", "type": "string"},
    {"name": "chainId", "type": "uint256"},
    {"name": "verifyingContract", "type": "address"},
]



from web3 import Web3


DEFAULT_IPFS_GATEWAY = "https://gateway.ipfscdn.io/ipfs/"

TW_IPFS_SERVER_URL = "https://upload.nftlabs.co"

PINATA_IPFS_URL = "https://api.pinata.cloud/pinning/pinFileToIPFS"

DEFAULT_API_KEY = "_gg7wSSi0KMBsdKnGVfHDueq6xMB9EkC"


def get_provider_for_network(network: str) -> Web3:
    """
    Returns a web3 provider instance for the given network.
    """

    rpc_url = ""
    if network == "mainnet":
        rpc_url = f"https://eth-mainnet.g.alchemy.com/v2/{DEFAULT_API_KEY}"
    elif network == "rinkeby":
        rpc_url = f"https://eth-rinkeby.g.alchemy.com/v2/{DEFAULT_API_KEY}"
    elif network == "goerli":
        rpc_url = f"https://eth-goerli.g.alchemy.com/v2/{DEFAULT_API_KEY}"
    elif network == "polygon":
        rpc_url = f"https://polygon-mainnet.g.alchemy.com/v2/{DEFAULT_API_KEY}"
    elif network == "mumbai":
        rpc_url = f"https://polygon-mumbai.g.alchemy.com/v2/{DEFAULT_API_KEY}"
    elif network == "fantom":
        rpc_url = "rpc.ftm.tools"
    elif network == "avalanche":
        rpc_url = "https://rpc.ankr.com/avalanche"
    else:
        if network.startswith("http"):
            rpc_url = network
        else:
            raise Exception(f"Unrecognized chain name or RPC url {network}")

    return Web3(Web3.HTTPProvider(rpc_url))



from typing import List, cast
from thirdweb.abi.token_erc20 import TokenERC20
from thirdweb.common.currency import (
    fetch_currency_metadata,
    fetch_currency_value,
    parse_units,
)
from thirdweb.constants.currency import ZERO_ADDRESS
from thirdweb.constants.role import Role, get_role_hash
from thirdweb.core.classes.contract_wrapper import ContractWrapper
from thirdweb.core.classes.base_contract import BaseContract
from thirdweb.core.classes.ipfs_storage import IpfsStorage
from thirdweb.types.currency import (
    Currency,
    CurrencyValue,
    Price,
    PriceWei,
    TokenAmount,
)
from web3.eth import TxReceipt


class ERC20(BaseContract[TokenERC20]):
    _storage: IpfsStorage

    def __init__(
        self,
        contract_wrapper: ContractWrapper,
        storage: IpfsStorage,
    ):
        super().__init__(contract_wrapper)
        self._storage = storage

    """
    READ FUNCTIONS
    """

    def get(self) -> Currency:
        """
        Get the token metadata including name, symbol, decimals, etc.

        ```python
        token = contract.get()
        print(token)
        ```

        :returns: token metadata
        """

        return fetch_currency_metadata(
            self._contract_wrapper.get_provider(), self.get_address()
        )

    def balance(self) -> CurrencyValue:
        """
        Get the token balance of the connected wallet.

        ```python
        balance = contract.balance()
        print(balance)
        ```

        :returns: balance of the connected wallet
        """

        return self.balance_of(self._contract_wrapper.get_signer_address())

    def balance_of(self, address: str) -> CurrencyValue:
        """
        Get the balance of the specified wallet

        ```python
        address = "{{wallet_address}}"
        balance = contract.balance_of(address)
        print(balance)
        ```

        :param address: wallet address to check the balance of
        :returns: balance of the specified wallet
        """

        return self._get_value(
            self._contract_wrapper._contract_abi.balance_of.call(address)
        )

    def total_supply(self) -> CurrencyValue:
        """
        Get the total minted supply of the token.

        :returns: total minted supply of the token
        """

        return self._get_value(self._contract_wrapper._contract_abi.total_supply.call())

    def allowance(self, spender: str) -> CurrencyValue:
        """
        Get a specific spenders allowance of this token for the connected wallet.

        ```python
        spender = "{{wallet_address}}"
        allowance = contract.allowance(spender)
        ```

        :param spender: wallet address to check the allowance of
        :returns: allowance of the connected wallet
        """

        return self.allowance_of(self._contract_wrapper.get_signer_address(), spender)

    def allowance_of(self, owner: str, spender: str) -> CurrencyValue:
        """
        Get the allowance of the specified spender for a specified owner.

        ```python
        # Address of the wallet who owns the funds
        address = "{{wallet_address}}"

        # Address of the wallet to check the token allowance
        spender = "0x..."

        allowance = contract.allowance_of(address, spender)
        print(allowance)
        ```

        :param owner: wallet address whose assets will be spent
        :param spender: wallet address to check the allowance of
        :returns: allowance of the specified spender for the specified owner
        """

        return self._get_value(
            self._contract_wrapper._contract_abi.allowance.call(owner, spender)
        )

    def is_transfer_restricted(self) -> bool:
        """
        Check whether transfer is restricted for tokens in this module.

        :returns: True if transfer is restricted, False otherwise
        """

        anyone_can_transfer = self._contract_wrapper._contract_abi.has_role.call(
            get_role_hash(Role.TRANSFER), ZERO_ADDRESS
        )

        return not anyone_can_transfer

    """
    WRITE FUNCTIONS
    """

    def transfer(self, to: str, amount: Price) -> TxReceipt:
        """
        Transfer a specified amount of tokens from the connected wallet to a specified address.

        ```python
        # Address to send tokens to
        to = "0x...

        # Amount of tokens to transfer
        amount = 0.1

        contract.transfer(to, amount)
        ```

        :param to: wallet address to transfer the tokens to
        :param amount: amount of tokens to transfer
        :returns: transaction receipt of the transfer
        """

        amount_with_decimals = parse_units(amount, self.get().decimals)
        return self._contract_wrapper.send_transaction(
            "transfer", [to, amount_with_decimals]
        )

    def transfer_from(self, fr: str, to: str, amount: Price) -> TxReceipt:
        """
        Transfer a specified amount of tokens from one specified address to another.

        ```python
        # Address to send tokens from
        fr = "{{wallet_address}}"

        # Address to send tokens to
        to = "0x..."

        # Amount of tokens to transfer
        amount = 0.1

        contract.transfer_from(fr, to, amount)
        ```

        :param fr: wallet address to transfer the tokens from
        :param to: wallet address to transfer the tokens to
        :param amount: amount of tokens to transfer
        :returns: transaction receipt of the transfer
        """

        amount_with_decimals = parse_units(amount, self.get().decimals)
        return self._contract_wrapper.send_transaction(
            "transfer_from", [fr, to, amount_with_decimals]
        )

    def set_allowance(self, spender: str, amount: Price) -> TxReceipt:
        """
        Sets the allowance of the specified wallet over the connected wallets funds to
        a specified amount.

        ```python
        spender = "0x..."
        amount = 100
        contract.set_allowance(spender, amount)
        ```

        :param spender: wallet address to set the allowance of
        :param amount: amount to set the allowance to
        :returns: transaction receipt of the allowance set
        """

        amount_with_decimals = parse_units(amount, self.get().decimals)
        return self._contract_wrapper.send_transaction(
            "approve", [spender, amount_with_decimals]
        )

    def transfer_batch(self, args: List[TokenAmount]):
        """
        Transfer tokens from the connected wallet to many wallets.

        ```python
        from thirdweb.types.currency import TokenAmount

        data = [
            TokenAmount("{{wallet_address}}", 0.1),
            TokenAmount("0x...", 0.2),
        ]

        contract.transfer_batch(data)
        ```

        :param args: list of token amounts and addressed to transfer to
        :returns: transaction receipt of the transfers
        """

        encoded = []
        interface = self._contract_wrapper.get_contract_interface()

        for arg in args:
            amount_with_decimals = self.normalize_amount(arg.amount)
            encoded.append(
                interface.encodeABI("transfer", [arg.to_address, amount_with_decimals])
            )

        return self._contract_wrapper.multi_call(encoded)

    def burn(self, amount: Price) -> TxReceipt:
        """
        Burn a specified amount of tokens from the connected wallet.

        ```python
        amount = 0.1
        contract.burn(amount)
        ```

        :param amount: amount of tokens to burn
        :returns: transaction receipt of the burn
        """

        amount_with_decimals = parse_units(amount, self.get().decimals)
        return self._contract_wrapper.send_transaction("burn", [amount_with_decimals])

    def burn_from(self, holder: str, amount: Price) -> TxReceipt:
        """
        Burn a specified amount of tokens from a specified wallet.

        ```python
        holder = "{{wallet_address}}"
        amount = 0.1
        contract.burn_from(holder, amount)
        ```

        :param holder: wallet address to burn the tokens from
        :param amount: amount of tokens to burn
        :returns: transaction receipt of the burn
        """

        amount_with_decimals = parse_units(amount, self.get().decimals)
        return self._contract_wrapper.send_transaction(
            "burn_from", [holder, amount_with_decimals]
        )

    """
    PROTECTED FUNCTIONS
    """

    def _get_value(self, value: PriceWei) -> CurrencyValue:
        return fetch_currency_value(
            self._contract_wrapper.get_provider(),
            self.get_address(),
            value,
        )

    def normalize_amount(self, amount: Price) -> PriceWei:
        decimals = self.get().decimals
        return parse_units(amount, decimals)



from dataclasses import dataclass
from typing import Any, Dict, Optional, Union


@dataclass
class NFTMetadataInput:
    """
    The metadata of an NFT to mint

    You can use the NFTMetadataInput.from_json(json) method to create
    an instance of this class from a dictionary.

    :param name: The name of the NFT
    :param description: The optional description of the NFT
    :param image: The optional image of the NFT
    :param external_url: The optional external URL of the NFT
    :param animation_url: The optional animation URL of the NFT
    :param background_color: The optional background color of the NFT
    :param properties: The optional properties of the NFT
    """

    name: str
    description: Optional[str] = None
    image: Optional[str] = None
    external_url: Optional[str] = None
    animation_url: Optional[str] = None
    background_color: Optional[str] = None
    properties: Optional[Dict[str, Any]] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "NFTMetadataInput":
        return NFTMetadataInput(
            json["name"],
            json.get("description"),
            json.get("image"),
            json.get("external_url"),
            json.get("animation_url"),
            json.get("background_color"),
            json.get("properties"),
        )

    def to_json(self) -> Dict[str, Any]:
        json: Dict[str, Any] = {}
        json["name"] = self.name
        if self.description is not None:
            json["description"] = self.description
        if self.image is not None:
            json["image"] = self.image
        if self.external_url is not None:
            json["external_url"] = self.external_url
        if self.animation_url is not None:
            json["animation_url"] = self.animation_url
        if self.background_color is not None:
            json["background_color"] = self.background_color
        if self.properties is not None:
            json["properties"] = self.properties

        return json


@dataclass
class NFTMetadata:
    id: int
    uri: str
    name: str
    description: Optional[str] = None
    image: Optional[str] = None
    external_url: Optional[str] = None
    animation_url: Optional[str] = None
    background_color: Optional[str] = None
    properties: Optional[Dict[Any, Any]] = None

    @staticmethod
    def from_json(json: Dict[str, Any]) -> "NFTMetadata":
        return NFTMetadata(
            json["id"],
            json["uri"],
            json["name"],
            json.get("description"),
            json.get("image"),
            json.get("external_url"),
            json.get("animation_url"),
            json.get("background_color"),
            json.get("properties"),
        )


@dataclass
class NFTMetadataOwner:
    metadata: NFTMetadata
    owner: str


@dataclass
class EditionMetadata:
    metadata: NFTMetadata
    supply: int


@dataclass
class EditionMetadataInput:
    """
    The metadata of an edition NFT to mint

    :param metadata: The metadata of the edition NFT
    :param supply: The supply of the edition NFT
    """

    metadata: Union[NFTMetadataInput, str]
    supply: int


@dataclass
class EditionMetadataOwner(EditionMetadata):
    owner: str
    quantity_owned: int


@dataclass
class QueryAllParams:
    start: int = 0
    count: int = 100



from dataclasses import dataclass
from typing import NewType

Price = float
PriceWei = int
Amount = Price


@dataclass
class Currency:
    name: str
    symbol: str
    decimals: int


@dataclass
class CurrencyValue(Currency):
    value: PriceWei
    display_value: Price


@dataclass
class TokenAmount:
    to_address: str
    amount: Price


@dataclass
class WrappedToken:
    address: str
    name: str
    symbol: str


@dataclass
class NativeToken:
    name: str
    symbol: str
    decimals: int
    wrapped: WrappedToken
